**********************************************************************
* Name:  BZ2
* Desc:  General purpose compresser compatible with BZ
* Author : Yann Collet
* Date : 12/2008
*
* Patterns:
* ^^^^^^^^^
* 1[MBITS][OBITS]  Match
* 0[NBITS]  Literals
*
* The bit count for offset is determined by the max possible
* offset, limited by traversed size.
* The 1-bit ID for matches is unnecessary after a literal since
* a match always follows.
**********************************************************************
ASSEMBLE

sMNEXT EQU 9 Match expected?
sGARB EQU 10

NBITS EQU 4 Literal length
NMASK EQU 2^(NBITS)-1
NBITS2 EQU 6
NMASK2 EQU 2^(NBITS2)-1
MBITS EQU 5 Match length
MMASK EQU 2^(MBITS)-1
MBITS2 EQU 7
MMASK2 EQU 2^(MBITS2)-1

RPL
::
  CK1NOLASTWD
  DUPTYPECSTR? ITE
    :: DUP ONE TWO SUB$ "BZ" EQUAL ;
    FALSE
  case

* True => Decompressor Code
CODEM

  ST=0.sGARB
*.reungz	

  SAVE
  C=DAT1.A                 % A = Object Pile 1
	 CD1EX                    % D1 ->stk1
	 RSTK=C                   % Save ->Stk1 into RSTK
	 D1+10                    % Skip Prolog & StrLength
	 A=DAT1.A
	 D1+4
	 LC(4) $5A42
	 P=3 ?C=A.WP -> .HeaderOK
	 P=0 LC(5)	=SETTYPEERR
  A=C.A LOAD PC=(A)        % EXIT avec error TYPE_ERR

*.HeaderOK	
  P=0
	 C=DAT1.A                 % C = ObSize
	 GOSBVL	=CREATETEMP
	 GONC	.ugzmemok

	 ?ST=1.sGARB ->	.ugzmerr
	 GOSUB	.PassGC
	 $(5) =DOCOL
	 $(5) =GARBAGE
	 $(5) =COLA
	 $(5) =DOCODE
	 G5 .UGZEND
	 ST=1.sGARB
	 GOTO	.reungz
*.PassGC	
  C=RSTK                   % C = Adress.of.DOCOL
  A=C.A LOAD PC=(A)        % Restore RPL & Start DOCOL
*.ugzmerr	
  GOVLNG	=GPMEMERR         % Exit with memory error

*.ugzmemok
		CD1EX R1=C.W	            % ->obend into R1
	 C=RSTK D1=C
	 C=DAT1.A                 % C : Objet Pile Niveau 1 (Source)
	 C+(5+5).A C+(5+4).A             % Skip Prolog, StrSize, Header, ObSize
	 A=C.A
		CD0EX R3=C.W             % D0 : ->Source / R3 = ->obstart
	 C=DAT0.A C+A.A R2=C.W	   % ->lits into R2
	 D0+5	                    % ->pats
	
	 A=R3.W DAT1=A.A	         % Swap ->stk1 by ->Ob
	 D1=A                     % D1 : Destination ->Ob

*.Uncompress_Loop
	 ST=0.sMNEXT
	 B=0.S       	            % UsedBits = 0

*.UPKLOOP	
{ GOSUB	.Decod
		?D#0.A -> .DecodMatch    % D#0 -> match
	 CD0EX CR2EX.A CD0EX      % D0 <-> R2 (->lits)
	 C=A.A GOSBVL	=MOVEDOWN   % Copy C.A Nibble from D0 to D1; Update D0 & D1
	 CD0EX CR2EX.A CD0EX      % R2 <-> D0
*.CheckEOF	
  AD1EX D1=A C=R1.A        % A=D1=->ob ; C=R1=SourceEnd
	 ?A<C.A UP }
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 	LOADRPL                  % End of Decoding  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

*.DecodMatch
	 CD1EX D1=C               % C = D1 ->ob
	 C-A.A	                   % ->src
	 CD0EX CDEX.A             % D0 ->src, D = ->pats, C = matchlength
	 A-C.A GOC .OverlapCopy   % if Offset-MatchLength < 0 --> Overlap
*.StreamCopy
	 GOSBVL	=MOVEDOWN
*.RestoreD0	
  C=D.A D0=C               % Restore D0 ->pats
	 GOTO	.CheckEOF

*.OverlapCopy	
		A+C.A A=0.P ?A#0.A ->	.StreamCopy  % if A=offset<16 --> OverlapCopy
  C-1.A
*.upkplp	
  { A=DAT0	B DAT1=A.1      % Copy 1 nibble
	 D0+1 D1+1 C-1.A UPNC }
	 GOC	.RestoreD0           % Always Carry
	 
*.UGZEND
  ENDCODE


**************************************************
* Compressor sysRPL header
**************************************************

  ZERO FOURTEEN BLANKIT
  ABUFF FOURTWO SEVEN
  "BZ2" BINT_131d
  CENTER$3x5 DROP

  DUP OSIZE SWAP
  GARBAGE
  SysTime 1LAMBIND


**************************************************
* Compressor ASM code
**************************************************

ASSEMBLE

HSHSIZE EQU #400
WINSIZE EQU #1000

RINGMEM EQU 3*(WINSIZE)
HASHMEM EQU 5*(HSHSIZE)+20
ALLMEM EQU (RINGMEM)+(HASHMEM)
MINADD EQU 100+(ALLMEM)

MINMAT EQU 6           Minimum match length

RPL

CODEM
  SAVE                 % D1 ->stk1
%		GOSBVL =DisableIntr
		CLRST
  A=0.W R0=A.W R1=A.W R2=A.W R3=A.W R4=A.W   % Init

  A=DAT1.A R3=A.A      % ->ob (temporary)
  D0=A GOSBVL =SKIPOB  % Skip object in D0, --> D0 = addr past object; Uses: A.A C.A P ST1 RSTK2; clears ST1, clears carry, P=0
  AD0EX R4=A.A         % ->obend into R4
  C=R3.A               % ->ob
  A-C.A                % obsize
  LC(5)10
  ?A>=C.A -> .argok    % ok if obsize>=10
*.argerr
  LC(5) =SETTYPEERR
*.gpevalc
  A=C.A LOAD PC=(A)    % EXIT with error C (TYPE_ERR here)

*.memerr
  GOSBVL =DispOn
  GOVLNG =GPMEMERR

*.argok                % A = obsize
	 C=A.A
	 CSR.A A+C.A
		LC(5)MINADD
	 C+A.A                % C = TmpSize
	 RSTK=C
	 
  GOSBVL =MAKE$N       % Creates character string in tempob area If not enough mem even after GC then memerr C.A = nibbles -> A=nibbles+5, B=nibbles+16 C=D1=addr of stack
                       % D0 = addr of body of $; R0 = addr of $ Not used: R1-R4
  A=R0.A GOSBVL =ASLW5 AD0EX R0=A.W   % R0.A=D0=addr of body of $; R0.5 = addr of $
  D1=A                 % ->out
  C=RSTK               % free
  GOSBVL =WIPEOUT      % Zeroes C.A nibbles at D1; D1 at end
  AD1EX LC(5)HASHMEM A-C.A   % HashMem at end
  R2=A.A               % ->buffer

%*********************************
%* R0 ->out ->$start
%* R1 ???
%* R2 ->buffer
%* R3 ->ob  bits
%* R4 ->obend (->obsave)
%*********************************
*.packnow
  C=R0.A D0=C          % D0 = R0 ->out
	 LC(4) $5A42          % LCSTR 'BZ'
  DAT0=C.4 D0+4
  A=R4.A               % ->obend
  C=R3.A               % ->ob
  A-C.A                % obsize
  DAT0=A.A D0+10
  CD0EX                % D0 = ->ob
  R0=C.A               % ->out
		A=R2.A D1=A D1+5
  C=R3.A DAT1=C.A      % save ->ob
  CD0EX D0=C           % C = D0
		GOTO .PackLoopCont

%*********************************
%* Compressor Loop
%*********************************
%* R0 ->out ->$start
%* R1 : Save Best MatchLength / Offset
%* R2 ->buffer
%* R3 ->ob  bits
%* R4 ->obend (->obsave)
%*********************************
*.Annunciator_and_On
  GOSUB .PackInfo_Abort?      % Annunciator flash
		GOTO .PackLoopCont

*.PackLoop
{ ?C=0.P -> .Annunciator_and_On
*.PackLoopCont
%* Save start address of scan
%* !!!! C = D0 !!!!
		RSTK=C               % D0 saved into RSTK
  GOSUB .FindMax       % Result : find a match, with litterals in front
                       % C = NibblesLeft

%---------------------------------------------------
% Changes for Greedy Match
%---------------------------------------------------

  AD0EX D0=A           % ->match
  C=RSTK               % C = D0 dpart
		A-C.A                % Skipped Nibbles
		
% IN : A=Literal Length, D.A=MatchLength, D.5 = Offset
  GOSUB .Encod
%* Pattern is out, update history
  ?D=0.A
  { D0+1                 % 1 in already
    D-2.A                % 1 + 1 for NC
%---------------------------------------------------
% end of Greedy Match Changes
%---------------------------------------------------

    { GOSUB .AddNew D0+1 D-1.A UPNC } % Rempli la Match_table de rfrences
		}
%* Prepare next loop
  CD0EX D0=C
  A=R4.A               % ->obend
  ?C>=A.A -> .PackNibs % Pos >= ->obend ==> On sort
  UP }


%*********************************
%* Append literal nibbles
%* according to packed directives
%*********************************
*.PackNibs             % Second pass, to add literals
		A=R2.A AD1EX D1+5
		C=DAT1.A R3=C.A      % restore ->ob

  C=R0.W
  A=R3.S ?A=0.S -> .outok!
  C+1.A R0=C.W         % R0+1.A
*.outok!
  GOSBVL =CSRW5
  C+(5+5).A C+(4+5).A  % Skip Prolog+StrSize; Skip Header+ObjSize
		D0=C A=C.A
  C=R0.W               % ->out
  C-A.A                % off to lit
  DAT0=C.A
  D0+5                 % ->pats

  C=R3.W D1=C          % ->ob
  ST=0.sMNEXT
	 B=0.S       	        % UsedBits = 0
*.NibsLoop
  GOSUB .Decod
  ?D#0.A -> .apppat
  C=R0.A CD1EX CD0EX R0=C.A  % D1=R0,D0=D1,R0=D0
  C=A.A
  GOSBVL =MOVEDOWN
  C=R0.A CD0EX CD1EX R0=C.A
  GONC .appcont        % Always Not Carry

*.apppat
  CD1EX C+D.A CD1EX

*.appcont
  C=R4.A               % ->obend
  AD1EX D1=A
  ?A<C.A -> .NibsLoop

%* Exit compressor
%* R0 ->out ->$start
*.PackDone
  C=R0.W D0=C
  GOSBVL =CSRW5 R0=C.W
  GOSBVL =Shrink$
%		GOSBVL =AllowIntr
  GOVLNG =GPOverWrR0Lp

%*********************************
%* Match Encoder
%*********************************
%* Output match
%* IN :
%* A[A]  = skipped
%* D[A]  = length
%* D[A1] = match loc
%* R0[A] = ->out
%* R3[S] = bits
%* Uses :
%* A[A] B[A] C[A] P
%*********************************
*.Encod
  ?A=0.A ->{ GOTO .outmat? }
% Output a Literal Length
  ST=1.sMNEXT
  LC(5)NMASK
  ?A>C.A -> .outlit2   %
  A+A.A P=((NBITS)+1) GOSUB .OutBits
  GOTO .outmat?
*.outlit2
  A-C.A B=A.A
  A=0.A P=((NBITS)+1) GOSUB .OutBits
  LC(5) NMASK2
  ?B>C.A -> .outlit5
  A=B.A P=(NBITS2) GOSUB .OutBits
  GOTO .outmat?
*.outlit5
  A=0.A P=(NBITS2) GOSUB .OutBits
  A=B.A GOSUB .OutA_A

%*********************************
*.outmat?
  ?D=0.A RTNYES
  D-("MINMAT"-1).A     % D = MatchLength - (Minmatch-1)
		LC(5)MMASK
  ?D>C.A -> .outmat2   % D > MaxMatchL ==> outmat2
  C=D.A A=C.A          % A = MatchLength -(Min-1)
  P=(MBITS)
		?ST=1.sMNEXT { A+A.A A+1.A P=((MBITS)+1) }  % Ajoute un Bit1
*.otm1
  GOSUB .OutBits
  GOTO .outoff
*.outmat2
  D-C.A                % D = MatchL - Max
  A=0.A P=(MBITS) ?ST=1.sMNEXT -> .otm3
  A+1.A P=((MBITS)+1)
*.otm3
  GOSUB .OutBits       % Outbits avec A=0 (-> Max ???)
  LC(5)MMASK2          % MMASK2 (4 x MMASK)
  ?D>C.A -> .outmat5   % If D=MatchL-Max1 > Max2 ==> outmat5
  C=D.A
  A=C.A                % A = D = MatchL - Max1
  P=(MBITS2) GOSUB .OutBits
  GOTO .outfxm
*.outmat5
  A=0.A P=(MBITS2) GOSUB .OutBits   % Outbits avec A=0 -> Max
  C=D.A A=C.A GOSUB .OutA_A   % OutA = OutBits with P=8, followed with another OutBits with P=12 => A.A entirely written
*.outfxm
  LC(5)MMASK
  D+C.A                % D = (MatchL - Max1) + Max1

*.outoff
  D+("MINMAT"-1).A     % Restore MatchLength
  ST=0.sMNEXT
  C=D.W GOSBVL =CSRW5  % ->matchloc
  AD0EX D0=A
  A-C.A                % offset

		C=R2.A C+5.A D1=C C=DAT1.A   % ->ob
  AD0EX C=A-C.A AD0EX  % maxoff
  B=C.A                % maxoff
  LC(5)$7FF
  P=12
{ ?B>C.A -> .OutBits   % if B(maxoff) > C(#7FF) then Outbits with P=12
  CSRB.A               % otherwise C.A / 2 (=#3FF) then P=11
  P-1 UPNC }           % BET (necessarily at some point)
%** Nb of bits to code offset depends on distance from beginning of file


%*********************************
% .OutA_A
% Write A.A in 2 steps (to Match .OutBits limitations)
%*********************************
% IN : A.A
% Uses : B.A P
%*********************************
*.OutA_A
  B=A.A P=8 GOSUB .OutBits
  A=B.A
  ASR.A ASR.A P=12    % then .OutBits

%*********************************
% OutBits
%*********************************
% IN :
%  A[A] Value to output
%  P = Nb Bits to output
%  R0[A] ->out
%  R3[S] bits used
% OUT :
%  Write Sequence at R0.A + R3.S.bits
%  Update R0.A & R3.S
% Uses :
%  A[A] A[S] C[A] C[S] P D1
%*********************************
*.OutBits
  A=R3.S               % A = R3.S = bits : determines how many bits are valid in current ->out, eg 0 <= bits <= 2
  C=P.15               % C.S = P
  C+A.S                % C.S = P + bits
  A=C.S                % A.S = P + bits
  C=0.A
  P=3 CPEX.15          % C.S = 3; P = P + bits
  CPEX.0               % C.0 = P + bits; P=0
  A&C.S                % A.S mod 4
  AR3EX.S              % R3.S = A.S = New.bits / A.S = previoux.bits
  CSRB.P CSRB.P        % C.0 / 4 (NbNibbles)
  AR0EX.A              % A.A = R0 ->out
  D1=A                 % D1 ->out
  A+C.A
  AR0EX.A              % Update R0 ->out
  A-1.S                % Si A.S=0 ==> NoShift
		SKIPC {
		 A+A.A A-1.S UPNC }  % A.A Shift bit left while A.S>0
  C=0.A
  C=DAT1.P             % Read current Nibble
  C!A.A                % OR
  DAT1=C.A             % write sequence (maybe more than necessary, but without consequence)
  RTN


%*********************************
%*.Decod	
%*********************************
%* Get directive from bitstream
%* In:	B.S = used_bits
%*     D0	= ->in
%*	    D1	= ->ob_ptr
%*	    R3[A]	= ->ob_start
%* Out:	A[A]=nlen D[A]=0
%*	     A[A]=offs D[A]=mlen
%* Update : B.S
%*********************************
*.Decod	
  ?ST=1.sMNEXT -> .dcmat   % This is a match
	 GOSUB .Get_1_or_5_Bits   % Get 1 bit
	 ?ABIT=1.0 -> .dcmat      % If Bit0=1 -> Match

% Decode Literal	Length
	 ST=1.sMNEXT              % Next time will be a match
		ASRB.B D0+1              % Get NBITS=4 bits			
  LC(5)NMASK A&C.A
	 D=0.A ?A#0.A RTNYES      % Nb of literals [1..15]
% Nb of literals > 15; need more bits
	 P=NBITS2 GOSUB	.GetPBits % Get NBITS2=6 bits
	 ?A#0.A -> .fxlit
	 GOSUB	.GetA_A
*.fxlit
  LC(5)NMASK A+C.A         % Renorm A (LiteralLength)
	 RTN

% Decode Match	Length
*.dcmat	
  ST=0.sMNEXT              % Check match bit next time
	 GOSUB .Get_1_or_5_Bits   % Get 5 bits
	 LC(5)MMASK A&C.A
  D0+1
	 ?A#0.A -> .decodeOffset  % Nb of matches [MINMATCH..MINMATCH+30]
% Nb of matches > MINMATCH+30; need more bits
	 P=MBITS2 GOSUB	.GetPBits % Get MBITS2=7 bits
	 ?A#0.A -> .fxmat
	 GOSUB	.GetA_A
*.fxmat	
  LC(5)MMASK A+C.A         % Renorm Matchlength

*.decodeOffset             % Read Offset
  C=A.A D=C.A	             % D = matchlength
	 D+((MINMAT)-1).A         % Renorm MatchLength
	 AD1EX D1=A               % A = D1
	 C=R3.A	                  % ->obstart
	 A-C.A	                   % A.A = distance from beginning
	 LC(5)$7FF P=12
	 ?A>C.A SKIPYES
{ CSRB.X P-1 ?A<=C.X UP }  % Less bits used for shorter distances
		C+C.A C+1.A              % mask
	 C=P.15
		GONC .GetBits            % Always NC

%*********************************************************************
% BitStream functions
%*********************************
*.GetA_A                   % Get full A.A using GetBits
  P=12 GOSUB	.GetPBits
 	B=A.X
 	P=8 GOSUB	.GetPBits
 	ASL.A ASL.A ASL.A
 	A=B.X
 	RTN

%*********************************
%* Get P bits
%*********************************
*.GetPBits                 % Valid for P in [6..12]
	 C=P.15
	 P=0 LC(5)32
	 A=C.S A-6.S
{ C+C.A A-1.S UPNC } C-1.A % building mask	
	
*.GetBits
  A=DAT0.A
		CBEX.S                   % C.S=BitsUsed
  B+C.S                    % B.S = New_BitsRead
	 C-1.S SKIPC              % if full nibble, no need to shift
{ ASRB.A C-1.S UPNC }      % "used bits" removed (from 1 to 3)

	 A&C.A                    % A.A => result (with required nb of bits)

		C=B.S                    % C.S = New_BitsRead
	 C=0.A
	 P=3 CPEX.15 CPEX.0       % C.S=3; C.A=New_BitsRead; P=0
	 B&C.S                    % B.S = BitsUsed
	
		CSRB.P CSRB.P            % NbFullNibblesUsed
	 AD0EX A+C.A AD0EX        % Update D0 (from +0 to +3)
	 RTN


*.Get_1_or_5_Bits
  A=DAT0.B                 % / 15
	 C=B.S                    % C.S=BitsUsed / 4
	 C-1.S SKIPC              % if full nibble, no need to shift / 4 (3-10)
{ ASRB.B C-1.S UPNC }      % "used bits" removed (from 1 to 3) / 8 4 (3-10)
		
	 C+4.S                    % C.S = 3 / 6
		B+1.S                    % B.S = New_BitsUsed / 4
	 C-B.S RTNNC    							   % (note : cannot use B-x.S for carry test, this opcode is buggy) / 4 (3-10)
	 D0+1 B=0.S               % Skip a full Nibble if New_BitsUsed=4 / 8 4
	 RTN                      % / 9


%*********************************
%* Match Finder (Main Loop)
%*********************************
%* Find next match causing a pack
%* In:
%* D0 ->pos
%* R2[A] ->buffer
%* R3[A] ->ob
%* R4[A] ->obend
%* Out:
%* D0 ->newpos
%* D[A] matchlen / D[A1] matchpos
%* Uses: A[W] B[A] C[W] D[W]
%*********************************
*.FindMax
{ GOSUB .AddNew        % Out: NotCarry if Match; A[A] = Address Match
  SKIPC {
		  GOSUB .ScanThis
    LC(5)(MINMAT)
    ?D>=C.A -> .Matchfound }   % D(MatchLength) must be > C(MinMatch) otherwise, nextmat
*.nxtmat
  AD0EX A+1.A D0=A     % D0+1
  C=R4.A ?A<C.A UP }   % On recommence tant que A=D0 < C=obend
  D=0.A RTN
*.Matchfound
  AD0EX D0=A           % A = D0
  C=R4.A C-A.A         % C = obend - currentpos = NibblesLeft
  ?D<=C.A RTNYES       % D(MatchLength) must be < C(NibblesLeft) otherwise D=C

		D=C.A LC(5)(MINMAT)  % D(MatchLength) corrected
		?D<C.A -> .nxtmat
  RTN

%*********************************
%* Ring Buffer Match Finder
%*********************************
%* Scan buffer for the longest
%* match for D0. Note that no
%* obend test is needed
%* In:
%* D0 ->pos
%* A[A] ->PrevMatch
%* R2[A] ->buffer
%* Out: D[A,A1] match <-----------
%* Uses: A[W] B[A] C[W] D[W] R3[A]
%*********************************	
*.ScanThis
  D=0.W D+(MINMAT-1).P
*.ScanforBetter	
		LC(5)WINSIZE-1
		AD0EX
		C=A-C.A R3=C.A       % R3 = D0-WinSize = LowLimit
		AD0EX
		?A<C.A RTNYES        % exit if match < LowerLimit
*.scanloop
  { D1=A GOSUB .Count  % Out : C.A
  ?D>C.A -> .nextmat   % D(PrevMatchCount) > C(MatchLength) => NextMatch
  CD1EX D=C.A CD1EX    % D = D1 = Adresse.Match
  P=9 DSL.WP DSL.WP DSL.WP DSL.WP DSL.WP P=0
  D=C.A                % D.W = C.A (MatchCount) | D.A (Adresse.Match)
%* Get next loc from buffer
*.nextmat
  CD1EX                % C = D1 / 8
		B=C.A                % / 7
		C=P.3 C=P.4          % C mod Winsize / 6 6
		A=R2.A               % ->buf / 11
		A-C.A A-C.A A-C.A    % ->slot / 7 7 7
		D1=A C=DAT1.3        % / 8 20
		?C=0.X RTNYES        % / (-16)
		B-C.A A=B.A          % / 7 7
		C=R3.A               % A = NextMatchPos; C = LowerLimit / 11
		?C<=A.A UP }         % / (-18)
		RTN                  % / 9

%*********************************
%* Match Length Counter
%*********************************
%* Compare nibbles
%* In: D0(CurrentPos) D1(MatchPos)
%* Out: C[A] = match length
%* Uses: A[W] B[A] C[W]
%*********************************
*.count0
	 C=0.A P=0 RTN

*.Count
		C=D.A CPEX.0 ?C=0.A { P=15 }
  A=DAT0.WP C=DAT1.WP
  ?C#A.WP -> .count0

		C=P.0 P=0
		B=0.A B=C.P B+1.B
		CD0EX C+B.A CD0EX
		CD1EX C+B.A CD1EX

{ A=DAT0.A C=DAT1.A
  ?C#A.A EXIT
		D0+5 D1+5 B+5.A UP }

  CD0EX C-B.A CD0EX            % Restore D0
  CD1EX C-B.A CD1EX            % Restore D1

  {     ?A#C.P EXIT            % P=0
    P=1 ?A#C.P -> .Add1        % P=1
    P=2 ?A#C.P -> .Add2        % P=2
    P=3 ?A#C.P -> .Add3        % P=3
       B+4.A EXITNC            % Note : max as A#C.A
*.Add3 B+3.A EXITNC
*.Add2 B+2.A EXITNC
*.Add1 B+1.A
    }
  C=B.A P=0
  RTN

%***************************************
%* Ring Buffer Addition
%***************************************
%* Add current location to ring
%* In: D0
%*     R2[A] ->buffer
%* Out: A[A] : Address Previous Match
%*      Carry: no previous match
%* Uses: A[5] C[5] D1 P
%***************************************
*.AddNew
%* Compute hash
  A=DAT0.6             % A = Sequence
  P=5 C=A.WP CSR.WP P=0
		CSR.A CSRB.A A+C.X
		CSR.A CSRB.X A+C.X
% Hash done; optimized for 10 bits
  LC(3)((HSHSIZE)-1)   % only 10 bits remain on C; no need for mask
  C&A.X                % C = Hash.CurrentSequence

  A=R2.A               % A = -> buffer
  A+C.A C+C.A C+C.A A+C.A    % 5*
  A+(5+5).A            % Skip MatchZero & ->ob
  D1=A                 % D1 -> Case Hash
  A=DAT1.A             % Most Recent Addr
  CD0EX D0=C           % C = D0 (CurrentSourcePos)
  DAT1=C.A             % New Most Recent Addr = Current Pos

		AD1EX                % Save MatchPos into D1
		B=C.A
		C=P.3 C=P.4          % C=D0 mod WinSize
		A=R2.A A-C.A A-C.A A-C.A   % A = ->slot
		AD1EX
		B-A.A C=B.X
		?B>C.A -> .TooBig
  DAT1=C.X             % ->slot(D0) = MatchPos
		RTN                  % No Carry : Match

*.TooBig
	 C=0.X DAT1=C.X RTN	  % Carry : No Match

%*********************************
%* Annunciator Flash
%*********************************
*.PackInfo_Abort?
  D1=(5) =ANNCTRL
  A=DAT1.B
  A+A.B
  ABIT=0.6
  ?A#0.B -> .inf00
  A+1.B
*.inf00
  ABIT=1.7
  DAT1=A.B

%* Abort Check
%*********************************
*.PackAbort?
		D1=(5) (=addrATTNFLG)+2
		A=DAT1.A D1=A
		A=DAT1.A ?A=0.A RTNYES
  A=R0.W GOSBVL =ASRW5 R0=A.W
		D0=A D0+10
  GOSBVL =Shrink$
  GOSBVL =DispOn
  LC(5) =ABORT
  GOTOL .gpevalc
%*********************************
ENDCODE

  SysTime 1GETLAM bit-
  ABND
  HXS>% % 8192 %/ %3 RNDXY UNROT
  DUP4UNROLL OSIZE
  ( $bz2 %time #len1 #len2 )
  UNCOERCE2 2DUP %CH
  %2 RNDXY a%>$
  "%CH: " !insert$
  4ROLL a%>$ "   Time: " !insert$
  !append$
  ZERO FOURTEEN BLANKIT
  ABUFF FOURTWO TEN 4ROLL BINT_131d
  CENTER$3x5 DROP
  %2 %/ a%>$ SWAP %2 %/ a%>$
  "Size: " !insert$
  " \8D " !append$ !insert$
  ABUFF FOURTWO ZERO 4ROLL BINT_131d
  CENTER$3x5 DROP
  SetDA1Temp
;

**********************************************************************
