***********************************************
* UBZ2
* Stand Alone decoder for BZ/BZ2
***********************************************


ASSEMBLE

sMNEXT	EQU 9	              Match expectation flag
sGARB	EQU 10               Garbage collector check flag

NBITS	EQU 4	               LiteralLength Bits
NMASK	EQU 2^(NBITS)-1
NBITS2	EQU 6
NMASK2	EQU 2^(NBITS2)-1
MBITS	EQU 5	               MatchLength Bits
MMASK	EQU 2^(MBITS)-1
MBITS2	EQU 7
MMASK2	EQU 2^(MBITS2)-1

MINMAT	EQU 6	              Minimum match length


RPL


CODEM

  ST=0.sGARB
*.reungz	

  SAVE
  C=DAT1.A                 % A = Object at stack level 1
	 CD1EX                    % D1 ->stk1
	 RSTK=C                   % Save ->Stk1 into RSTK
	 D1+10                    % Skip Prolog & StrLength
	 A=DAT1.A
	 D1+4
	 LC(4) $5A42              % "BZ"
	 P=3 ?C=A.WP -> .HeaderOK
	 P=0 LC(5)	=SETTYPEERR
  A=C.A LOAD PC=(A)        % EXIT with error TYPE_ERR

*.HeaderOK	
  P=0
	 C=DAT1.A                 % C = ObSize
	 GOSBVL	=CREATETEMP       % Allocates C.A nibbles; carry if notenoughmemory; ->D0=bottom,D1=topofarea,->B.A=C.A=@D1=offset
	 GONC	.ugzmemok

	 ?ST=1.sGARB ->	.ugzmerr
	 GOSUB	.PassGC
	 $(5) =DOCOL
	 $(5) =GARBAGE
	 $(5) =COLA
	 $(5) =DOCODE
	 G5 .UGZEND
	 ST=1.sGARB
	 GOTO	.reungz
*.PassGC	
  C=RSTK                   % C = Adress.of.DOCOL
  A=C.A LOAD PC=(A)        % Restore RPL & Start DOCOL
*.ugzmerr	
  GOVLNG	=GPMEMERR         % "Not enough memory" event detected

*.ugzmemok
		CD1EX R1=C.W	            % ->obend into R1
	 C=RSTK D1=C
	 C=DAT1.A                 % C : Stack level 1 (Source)
	 C+(5+5).A C+(5+4).A      % Skip Prolog, StrSize, Header, ObSize
	 A=C.A
		CD0EX R3=C.W             % D0 : ->Source / R3 = ->obstart
	 C=DAT0.A C+A.A R2=C.W	   % ->lits into R2
	 D0+5	                    % ->pats
	
	 A=R3.W DAT1=A.A	         % Replace ->stk1 by ->Ob
	 D1=A                     % D1 : Destination ->Ob

	 ST=0.sMNEXT
	 B=0.S       	            % UsedBits = 0

%**************************************
%* Main Decoding Loop
%**************************************
%* Get directive instructions
%* Copy nibbles according to directives
%* In:	B.S = used_bits
%*     D0	= ->in
%*	    D1	= ->ob_ptr
%*	    R1[A]	= ->ob_end
%*	    R2[A]	= ->lits
%*	    R3[A]	= ->ob_start
%* Uses : A.W C.A C.S D.A P ST.sMNEXT
%**************************************
{ GOSUB	.Decod             % Get Directive (D.A=0 => Literals; D.A>0 => Match)
		?D#0.A -> .MatchCopy     % D#0 -> This is a match
%* Literals Copy
	 CD0EX CR2EX.A CD0EX      % D0 <-> R2 (->lits)
	 C=A.A GOSBVL	=MOVEDOWN   % Copy C.A Nibble from D0 to D1; Update D0 & D1 ; Used: A.W C.A P
	 CD0EX CR2EX.A CD0EX      % R2 <-> D0
*.CheckEOF	
  AD1EX D1=A C=R1.A        % A=D1=->ob ; C=R1=SourceEnd
	 ?A<C.A UP }
	 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 	LOADRPL                  % End of Decoding  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%*********************************
%*.MatchCopy
%*********************************
%* Copy Nibbles from past position
%* In:	A[A]=offs D[A]=mlen
%*	    D1	= ->obptr
%* Update:	D1
%* Uses : A.W C.A D.A P
%*********************************
*.MatchCopy
	 CD1EX D1=C               % C = D1 ->ob
	 C-A.A	                   % ->src
	 CD0EX CDEX.A             % D0 ->src, D = ->bitstream, C = matchlength
	 A-C.A GOC .OverlapCopy   % If Offset-MatchLength < 0 --> Overlap
*.StreamCopy
	 GOSBVL	=MOVEDOWN         % Copy C.A Nibble from D0 to D1; Update D0 & D1 ; Used: A.W C.A P
*.RestoreD0	
  C=D.A D0=C               % Restore D0 ->pats
	 GOTO	.CheckEOF

*.OverlapCopy	
		A+C.A A=0.P ?A#0.A ->	.StreamCopy  % if A=offset>=16 --> StreamCopy
  C-1.A
  { A=DAT0.B DAT1=A.1      % Copy nibble one by one
	 D0+1 D1+1 C-1.A UPNC }
	 GOC	.RestoreD0           % Always Carry


%*********************************
%*.Decod	
%*********************************
%* Get directive from bitstream
%* In:	B.S = used_bits
%*     D0	= ->in
%*	    D1	= ->ob_ptr
%*	    R3[A]	= ->ob_start
%* Out:	A[A]=nlen D[A]=0
%*	     A[A]=offs D[A]=mlen
%* Update : B.S
%*********************************
*.Decod	
  ?ST=1.sMNEXT -> .dcmat   % This is a match
	 GOSUB .Get_1_or_5_Bits   % Get 1 bit
	 ?ABIT=1.0 -> .dcmat      % If Bit0=1 -> Match

% Decode Literal	Length
	 ST=1.sMNEXT              % Next time will be a match
		ASRB.B D0+1              % Get NBITS=4 bits			
  LC(5)NMASK A&C.A
	 D=0.A ?A#0.A RTNYES      % Nb of literals [1..15]
% Nb of literals > 15; need more bits
	 P=NBITS2 GOSUB	.GetPBits % Get NBITS2=6 bits
	 ?A#0.A -> .fxlit
	 GOSUB	.GetA_A
*.fxlit
  LC(5)NMASK A+C.A         % Renorm A (LiteralLength)
	 RTN

% Decode Match	Length
*.dcmat	
  ST=0.sMNEXT              % Check match bit next time
	 GOSUB .Get_1_or_5_Bits   % Get 5 bits
	 LC(5)MMASK A&C.A
  D0+1
	 ?A#0.A -> .decodeOffset  % Nb of matches [MINMATCH..MINMATCH+30]
% Nb of matches > MINMATCH+30; need more bits
	 P=MBITS2 GOSUB	.GetPBits % Get MBITS2=7 bits
	 ?A#0.A -> .fxmat
	 GOSUB	.GetA_A
*.fxmat	
  LC(5)MMASK A+C.A         % Renorm Matchlength
  
*.decodeOffset             % Read Offset
  C=A.A D=C.A	             % D = matchlength
	 D+((MINMAT)-1).A         % Renorm MatchLength
	 AD1EX D1=A               % A = D1
	 C=R3.A	                  % ->obstart
	 A-C.A	                   % A.A = distance from beginning
	 LC(5)$7FF P=12
	 ?A>C.A SKIPYES
{ CSRB.X P-1 ?A<=C.X UP }  % Less bits used for shorter distances
		C+C.A C+1.A              % mask
	 C=P.15
		GONC .GetBits            % Always NC

%*********************************************************************
% BitStream functions
%*********************************
*.GetA_A                   % Get full A.A using GetBits
  P=12 GOSUB	.GetPBits
 	B=A.X
 	P=8 GOSUB	.GetPBits
 	ASL.A ASL.A ASL.A
 	A=B.X
 	RTN

%*********************************
%* Get P bits
%*********************************
*.GetPBits                 % Valid for P in [6..12]
	 C=P.15
	 P=0 LC(5)32
	 A=C.S A-6.S
{ C+C.A A-1.S UPNC } C-1.A % building mask	
	 
*.GetBits
  A=DAT0.A
		CBEX.S                   % C.S=BitsUsed
  B+C.S                    % B.S = New_BitsRead
	 C-1.S SKIPC              % if full nibble, no need to shift
{ ASRB.A C-1.S UPNC }      % "used bits" removed (from 1 to 3)

	 A&C.A                    % A.A => result (with required nb of bits)

		C=B.S                    % C.S = New_BitsRead
	 C=0.A
	 P=3 CPEX.15 CPEX.0       % C.S=3; C.A=New_BitsRead; P=0
	 B&C.S                    % B.S = BitsUsed
	
		CSRB.P CSRB.P            % NbFullNibblesUsed
	 AD0EX A+C.A AD0EX        % Update D0 (from +0 to +3)
	 RTN


*.Get_1_or_5_Bits
  A=DAT0.B                 % / 15
	 C=B.S                    % C.S=BitsUsed / 4
	 C-1.S SKIPC              % if full nibble, no need to shift / 4 (3-10)
{ ASRB.B C-1.S UPNC }      % "used bits" removed (from 1 to 3) / 8 4 (3-10)
		
	 C+4.S                    % C.S = 3 / 6
		B+1.S                    % B.S = New_BitsUsed / 4
	 C-B.S RTNNC    							   % (note : cannot use B-x.S for carry test, this opcode is buggy) / 4 (3-10)
	 D0+1 B=0.S               % Skip a full Nibble if New_BitsUsed=4 / 8 4
	 RTN                      % / 9

*.UGZEND
ENDCODE




